package PVE::Storage;

use strict;
use POSIX;
use IO::Select;
use IO::Dir;
use IO::File;
use Fcntl ':flock';
use File::stat;
use File::Basename;
use File::Path;
use IPC::Open2;
use IPC::Open3;
use Filesys::Df;
use Cwd 'abs_path';
use Getopt::Long qw(GetOptionsFromArray);
use Socket;
use Digest::SHA1;

my $ISCSIADM = '/usr/bin/iscsiadm';
my $UDEVADM = '/sbin/udevadm';

$ISCSIADM = undef if ! -X $ISCSIADM;

my $storageconfig = "/etc/pve/storage.cfg";

my $lockfile = "/etc/pve/.lock-storage.cfg";

# we need the cluster configuration for locking operations

sub read_hostname {

    my $filename = "/etc/hostname";

    my $hostname = 'localhost';

    if (my $fh = IO::File->new ($filename, 'r')) {
	$hostname = <$fh>;
	chomp $hostname;
	close $fh;
    }

    return $hostname;
}

sub read_cluster_config {
    my ($filename, $fh) = @_;

    my $line;

    my $cinfo;

    $cinfo->{nodes} = [];

    my $maxcid = 0;
    my $level = 0;

    while (defined ($line = <$fh>)) {
	chomp $line;

	next if $line =~ m/^\s*$/; # skip empty lines

	if ($line =~ m/^maxcid\s+(\d+)\s*$/i) {
	    $maxcid = $1 > $maxcid ? $1 : $maxcid;
	    next;
	}
		
	if ($line =~ m/^(master|node)\s+(\d+)\s+\{\s*$/i) {
	    $level++;
	    my ($t, $cid) = (lc $1, $2);
		
	    $maxcid = $cid > $maxcid ? $cid : $maxcid;

	    my $res = { 
		role => $t eq 'master' ? 'M' : 'N',
		cid => $cid
		};

	    while (defined ($line = <$fh>)) {
		chomp $line;
		next if $line =~ m/^\s*$/; # skip empty lines
		if ($line =~ m/^\}\s*$/) {
		    $level--;
		    last;
		}
	    
		if ($line =~ m/^\s*(\S+)\s*:\s*(\S+)\s*$/) {
		    my ($n, $v) = (lc $1, $2);

		    # fixme: do syntax checks
		    if ($n eq 'ip') {
			$res->{$n} = $v;
		    } elsif ($n eq 'name') {
			$res->{$n} = $v;
		    } elsif ($n eq 'hostrsapubkey') {
			$res->{$n} = $v;
		    } elsif ($n eq 'rootrsapubkey') {
			$res->{$n} = $v;
		    } else {
			die "syntax error in configuration file\n";
		    }
		} else {
		    die "syntax error in configuration file\n";
		}
	    }

	    die "missing ip address for node '$cid'\n" if !$res->{ip};
	    die "missing name for node '$cid'\n" if !$res->{name};
	    die "missing host RSA key for node '$cid'\n" if !$res->{hostrsapubkey};
	    die "missing user RSA key for node '$cid'\n" if !$res->{rootrsapubkey};

	    push @{$cinfo->{nodes}}, $res;
		    
	    if ($res->{role} eq 'M') {
		$cinfo->{master} = $res;
	    } 
	} else {
	    die "syntax error in configuration file\n";
	}
    }

    die "syntax error in configuration file\n" if $level;

    $cinfo->{maxcid} = $maxcid;

    foreach my $ni (@{$cinfo->{nodes}}) {
	$cinfo->{"CID_$ni->{cid}"} = $ni;
    }

    my $hostname = read_hostname();

    $cinfo->{local} = {
	role => '-',
	cid => 0,
	ip => '127.0.0.1',
	name => $hostname,
    };

    foreach my $ni (@{$cinfo->{nodes}}) {
	if ($ni->{name} eq $hostname) {
	    $cinfo->{local} = $ni;
	    last;
	}
    }
 
    return $cinfo;
}

# generic utility function

sub check_iscsi_support {
    my $noerr = shift;

    if (!$ISCSIADM) {
	my $msg = "no iscsi support - please install open-iscsi";
	if ($noerr) {
	    warn "warning: $msg\n";
	    return 0;
	}

	die "error: $msg\n";
    }

    return 1;
}

sub link_points_to {
    my ($src, $dest) = @_;

    my $st1 = stat ($src);
    my $st2 = stat ($dest);

    return 0 if !$st1 || !$st2;

    return $st1->ino == $st2->ino && $st1->dev == $st2->dev;
}

sub find_stable_path {
    my ($stabledir, $bdev) = @_;

    my $dh = IO::Dir->new ($stabledir);
    if ($dh) {
	while (defined(my $tmp = $dh->read)) {
	    my $path = "$stabledir/$tmp";
	    if (link_points_to ($path, $bdev)) {
		return wantarray ? ($path, $tmp) : $path;
	    }
	}

	$dh->close;
    }

    return wantarray ? () : undef;
}

sub find_dev_by_id {
    my $bdev = shift;

    my ($full_path, $name) = find_stable_path ("/dev/disk/by-id", $bdev);

    return $name;
}

sub find_dev_by_path {
    my $bdev = shift;

    my ($full_path, $name) = find_stable_path ("/dev/disk/by-path", $bdev);

    return $name;
}

sub dir_glob_regex {
    my ($dir, $regex) = @_;

    my $dh = IO::Dir->new ($dir);
    return wantarray ? () : undef if !$dh;
  
    while (defined(my $tmp = $dh->read)) { 
	if (my @res = $tmp =~ m/^($regex)$/) {
	    $dh->close;
	    return wantarray ? @res : $tmp;
	}
    }
    $dh->close;

    return wantarray ? () : undef;
}

sub dir_glob_foreach {
    my ($dir, $regex, $func) = @_;

    my $dh = IO::Dir->new ($dir);
    if (defined $dh) {
	while (defined(my $tmp = $dh->read)) {
	    if (my @res = $tmp =~ m/^($regex)$/) {
		&$func (@res);
	    }
	}
    } 
}

sub file_read_firstline {
    my ($filename) = @_;

    my $fh = IO::File->new ($filename, "r");
    return undef if !$fh;
    my $res = <$fh>;
    chomp $res;
    $fh->close;
    return $res;
}

sub run_command {
    my ($cmd, %param) = @_;

    my $reader = IO::File->new();
    my $writer = IO::File->new();
    my $error  = IO::File->new();

    $cmd = [ $cmd ] if !ref($cmd);

    my $cmdstr = join (' ', @$cmd);

    my $timeout;
    my $input;
    my $outfunc;
    my $errfunc;

    foreach my $p (keys %param) {
	if ($p eq 'timeout') {
	    $timeout = $param{$p};
	} elsif ($p eq 'input') {
	    $input = $param{$p};
	} elsif ($p eq 'outfunc') {
	    $outfunc = $param{$p};
	} elsif ($p eq 'errfunc') {
	    $errfunc = $param{$p};
	} else {
	    die "got unknown parameter '$p' for run_command\n";
	}
    }

    # try to avoid locale related issues/warnings
    my $lang = $param{lang} || 'C'; 
 
    my $orig_pid = $$;

    my $pid;
    eval {
	local $ENV{LANG} = $lang;

	# suppress LVM warnings like: "File descriptor 3 left open";
	local $ENV{LVM_SUPPRESS_FD_WARNINGS} = "1";

	$pid = open3 ($writer, $reader, $error, @$cmd) || die $!;
    };

    my $err = $@;

    # catch exec errors
    if ($orig_pid != $$) {
	warn "ERROR: $err";
	POSIX::_exit (1); 
	kill ('KILL', $$); 
    }

    die $err if $err;

    print $writer $input if defined $input;
    close $writer;

    my $select = new IO::Select;
    $select->add ($reader);
    $select->add ($error);

    my $outlog = '';
    my $errlog = '';

    while ($select->count) {
	my @handles = $select->can_read ($timeout);

	if (defined ($timeout) && (scalar (@handles) == 0)) {
	    kill (9, $pid);
	    waitpid ($pid, 0);
	    die "command '$cmdstr' failed: timeout";
	}

	foreach my $h (@handles) {
	    my $buf = '';
	    my $count = sysread ($h, $buf, 4096);
	    if (!defined ($count)) {
		my $err = $!;
		kill (9, $pid);
		waitpid ($pid, 0);
		die "command '$cmdstr' failed: $err";
	    }
	    $select->remove ($h) if !$count;
	    if ($h eq $reader) {
		if ($outfunc) {
		    eval {
			$outlog .= $buf;
			while ($outlog =~ s/^([^\010\r\n]*)(\r|\n|(\010)+|\r\n)//s) {
			    my $line = $1;
			    &$outfunc ($line);
			}
		    };
		    my $err = $@;
		    if ($err) {
			kill (9, $pid);
			waitpid ($pid, 0);
			die $err;
		    }
		} else {
		    print $buf;
		    *STDOUT->flush();
		}
	    } elsif ($h eq $error) {
		if ($errfunc) {
		    eval {
			$errlog .= $buf;
			while ($errlog =~ s/^([^\010\r\n]*)(\r|\n|(\010)+|\r\n)//s) {
			    my $line = $1;
			    &$errfunc ($line);
			}
		    };
		    my $err = $@;
		    if ($err) {
			kill (9, $pid);
			waitpid ($pid, 0);
			die $err;
		    }
		} else {
		    print STDERR $buf;
		    *STDERR->flush();
		}
	    }
	}
    }

    &$outfunc($outlog) if $outfunc && $outlog;
    &$errfunc($errlog) if $errfunc && $errlog;

    my $rv = waitpid ($pid, 0);
    my $ec = ($? >> 8);

    die "command '$cmdstr' failed with exit code $ec\n" if $ec;
}

sub read_proc_mounts {
    
    local $/; # enable slurp mode
    
    my $data = "";
    if (my $fd = IO::File->new ("/proc/mounts", "r")) {
	$data = <$fd>;
	close ($fd);
    }

    return $data;
}

#  PVE::Storage utility functions

sub lock_config {
    my ($code, @param) = @_;

    my $timeout = 60;

    my $fh;

    my $res;

    eval {

	local $SIG{ALRM} = sub { die "got timeout\n"; };

	alarm ($timeout);

	$fh = new IO::File (">>$lockfile") ||
	    die "can't open lock for '$lockfile' - $!\n";

	if (!flock ($fh, LOCK_EX|LOCK_NB)) {
	    print STDERR "trying to aquire lock...";
	    if (!flock ($fh, LOCK_EX)) {
		print STDERR " failed\n";
		die "can't aquire lock for '$lockfile' - $!\n";
	    }
	    print STDERR " OK\n";
	}
	alarm (0);

	$res = &$code (@param);
    };

    my $err = $@;

    alarm (0);

    close ($fh) if $fh;

    if ($err) {
	$@ = $err;
	return undef;
    }

    return $res;
}

my $confvars = {
    path => 'path',
    shared => 'bool',
    disable => 'bool',
    format => 'format',
    content => 'content',
    server => 'server',
    export => 'path',
    vgname => 'vgname',
    base   => 'volume',
    portal => 'portal',
    target => 'target',
    options => 'string',
};

my $required_config = {
    dir => ['path'],
    nfs => ['path', 'server', 'export'],
    lvm => ['vgname'],
    iscsi => ['portal', 'target'],
};

my $fixed_config = {
    dir => ['path'],
    nfs => ['path', 'server', 'export'],
    lvm => ['vgname', 'base'],
    iscsi => ['portal', 'target'],
};

my $default_config = {
    dir => {
	path => 1,
	shared => 0,
	disable => 0,
	content => [ { images => 1, rootdir => 1, vztmpl => 1, iso => 1, backup => 1, none => 1 },
		     { images => 1,  rootdir => 1 }],
	format => [ { raw => 1, qcow2 => 1, vmdk => 1 } , 'raw' ],
    },

    nfs => {
	path => 1,
	disable => 0,
        server => 1,
        export => 1,
        options => 0,
	content => [ { images => 1, iso => 1, backup => 1},
		     { images => 1 }],
	format => [ { raw => 1, qcow2 => 1, vmdk => 1 } , 'raw' ],
    },

    lvm => {
	vgname => 1,
	shared => 0,
	disable => 0,
	content => [ {images => 1}, { images => 1 }],
        base => 1,
    },

    iscsi => {
        portal => 1,
        target => 1,
	disable => 0,
	content => [ {images => 1, none => 1}, { images => 1 }],
    },
};

sub valid_content_types {
    my ($stype) = @_;

    my $def = $default_config->{$stype};

    return {} if !$def;

    return $def->{content}->[0];
}

sub check_type {
    my ($stype, $ct, $key, $value, $storeid) = @_;

    my $def = $default_config->{$stype};

    return undef if !defined($def);
    return undef if !defined($def->{$key});

    return undef if !defined ($value);
    return undef if $value =~ m/[\n\r]/;

    if ($ct eq 'bool') {
	return 1 if ($value eq '1') || ($value =~ m/^(on|yes|true)$/i); 
	return 0 if ($value eq '0') || ($value =~ m/^(off|no|false)$/i); 
    } elsif ($ct eq 'path') {
	return undef if $value !~ m|^/.+|; # we need absolute paths
	return $value;
    } elsif ($ct eq 'server') {
	return undef if $value =~ m/^\s*$/; # no emtpy strings
	return undef if $value =~ m/\s/; # no white space
	return $value; # fixme
    } elsif ($ct eq 'vgname') {
	return undef if !parse_lvm_name ($value, 1);
	return $value;
    } elsif ($ct eq 'portal') {
	return undef if $value !~ m/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}(:\d+)?$/;
	return $value;
    } elsif ($ct eq 'target') {
	return undef if $value =~ m/^\s*$/; # no emtpy strings
	return undef if $value =~ m/\s/; # no white space
	return $value;
    } elsif ($ct eq 'string') {
	return $value;
    } elsif ($ct eq 'format') {
	my $valid_formats = $def->{format}->[0];

	return undef if !$valid_formats->{$value};

	return $value;

    } elsif ($ct eq 'content') {
	my $valid_content = $def->{content}->[0];
	
	$value =~ s/[,;]/ /g;
	$value =~ s/\s+/ /g;

	my $res = {};

	my @ca = split (/\s/, $value);
	foreach my $c (@ca) {
	    return undef if !$valid_content->{$c};
	    $res->{$c} = 1;
	} 

	# only local storage may have several content types
	if ($res->{none} || !($storeid && $storeid eq 'local')) {
	    return undef if scalar (keys %$res) > 1;
	}

	# no backup to local storage
	return undef if $storeid && $storeid eq 'local' && $res->{backup};

	return $res;	
    } elsif ($ct eq 'volume') {
	return $value if parse_volume_id ($value, 1);
    }

    return undef;
}

sub parse_config {
    my ($filename, $fh) = @_;

    my $ids = {};

    my $sha1 = Digest::SHA1->new;

    my $pri = 0;
    while ($fh && defined (my $line = <$fh>)) {
	$sha1->add ($line); # compute digest

	chomp $line;

	next if $line =~ m/^\#/;
	next if $line =~ m/^\s*$/;

	if ($line =~ m/^(\S+):\s*(\S+)\s*$/) {
	    my $storeid = $2;
	    my $type = $1;
	    my $ignore = 0;

	    if (!parse_storage_id ($storeid, 1)) {
		$ignore = 1;
		warn "ignoring storage '$storeid' - (illegal characters)\n";
	    } elsif (!$default_config->{$type}) {
		$ignore = 1;
		warn "ignoring storage '$storeid' (unsupported type '$type')\n";
	    } else {
		$ids->{$storeid}->{type} = $type;
		$ids->{$storeid}->{priority} = $pri++;
	    }

	    while (defined ($line = <$fh>)) {
		next if $line =~ m/^\#/;
		last if $line =~ m/^\s*$/;

		next if $ignore; # skip

		if ($line =~ m/^\s+(\S+)(\s+(.*\S))?\s*$/) {
		    my ($k, $v) = ($1, $3);
		    if (my $ct = $confvars->{$k}) {
			$v = 1 if $ct eq 'bool' && !defined($v);
			if (defined($v) &&
			    defined (my $res = check_type ($type, $ct, $k, $v, $storeid))) {
			    $ids->{$storeid}->{$k} = $res;
			} else {
			    warn "storage '$storeid' - unable to parse value of '$k'\n";
			}
		    } else {
			warn "storage '$storeid' - unable to parse value of '$k'\n";
		    }

		} else {
		    warn "storage '$storeid' - ignore config line: $line\n";
		}
	    }
	} else {
	    warn "ignore config line: $line\n";
	}
    }

    # make sure we have a reasonable 'local:' storage
    # openvz expects things to be there
    if (!$ids->{local} || $ids->{local}->{type} ne 'dir' ||
	$ids->{local}->{path} ne '/var/lib/vz') {
	$ids->{local} = {
	    type => 'dir',
	    priority => $pri++,
	    path => '/var/lib/vz',    
	    content => { images => 1, rootdir => 1, vztmpl => 1, iso => 1},
	};
    }

    # we always need this for OpenVZ
    $ids->{local}->{content}->{rootdir} = 1;
    $ids->{local}->{content}->{vztmpl} = 1;
    delete ($ids->{local}->{disable});

    foreach my $storeid (keys %$ids) {
	my $d = $ids->{$storeid};

	my $req_keys = $required_config->{$d->{type}};
	foreach my $k (@$req_keys) {
	    if (!defined ($d->{$k})) {
		warn "ignoring storage '$storeid' - missing value " .
		    "for required option '$k'\n";
		delete $ids->{$storeid};
		next;		
	    }
	}

	my $def = $default_config->{$d->{type}};

	if ($def->{content}) {
	    $d->{content} = $def->{content}->[1] if !$d->{content};
	}

	if ($d->{type} eq 'iscsi' || $d->{type} eq 'nfs') {
	    $d->{shared} = 1;
	}
    }

    my $digest = $sha1->hexdigest;

    my $cfg = { ids => $ids, digest => $digest};

    return $cfg;
}

sub parse_options {
    my ($storeid, $stype, $param, $create) = @_;

    my $settings = { type => $stype };

    die "unknown storage type '$stype'\n"
	if !$default_config->{$stype};

    my $errmsg;

    my $get_option_func = sub {
	my ($opt, $value) = @_;

	my $ct = $confvars->{$opt};
	if (defined($value) && defined (my $res = check_type ($stype, $ct, $opt, $value, $storeid))) {
	    $settings->{$opt} = $res;
	} else {
	    $errmsg = "unable to parse value for '$opt'";
	    die("!FINISH"); # abort GetOption
 	}
    };

    my $getopt;
    foreach my $opt (keys %{$default_config->{$stype}}) {
	$getopt->{"$opt=s"} = $get_option_func;
    }

    if (!GetOptionsFromArray ($param, %$getopt) || $errmsg) {
	if ($errmsg) {
	    die "$errmsg\n";
	} else {
	    die "unable to parse options\n";
	}
    }

    if ($create) {
	my $req_keys = $required_config->{$stype};
	foreach my $k (@$req_keys) {

	    if ($stype eq 'nfs' && !$settings->{path}) {
		$settings->{path} = "/mnt/pve/$storeid";
	    }

	    # check if we have a value for all required options
	    die "missing value for required option '$k'\n"
		if !defined ($settings->{$k});
	}
    } else {
	my $fixed_keys = $fixed_config->{$stype};
	foreach my $k (@$fixed_keys) {

	    # only allow to change non-fixed values
	    die "can't change value for option '$k'\n"
		if defined ($settings->{$k});
	}
    }

    return $settings;

}

sub load_config {

    my $cfg = {};

    my $fh = IO::File->new ($storageconfig, 'r');
    $cfg = parse_config ($storageconfig, $fh); # returns defaults if !$fh;
    $fh->close() if $fh;

    return $cfg;
}

sub load_cluster_config {

    my $cfg;

    my $filename = "/etc/pve/cluster.cfg";

    if (my $fh = IO::File->new ($filename, 'r')) {
	$cfg = read_cluster_config ($filename, $fh);
	$fh->close();
    }

    return $cfg;
}

sub fork_command_pipe {
    my ($cmd) = @_;
    
    my $reader = IO::File->new();
    my $writer = IO::File->new();

    my $orig_pid = $$;

    my $cpid;

    eval { $cpid = open2 ($reader, $writer, @$cmd); };

    my $err = $@;

    # catch exec errors
    if ($orig_pid != $$) {
	logmsg ('err', "can't fork command pipe\n");
	POSIX::_exit (1); 
	kill ('KILL', $$); 
    }

    die $err if $err;

    return { writer => $writer, reader => $reader, pid => $cpid };
}

sub finish_command_pipe { 
    my $cmdpipe = shift;

    my $writer = $cmdpipe->{writer};
    my $reader = $cmdpipe->{reader};

    $writer->close();
    $reader->close();

    my $cpid = $cmdpipe->{pid};

    kill (15, $cpid) if kill (0, $cpid);

    waitpid $cpid, 0;
}

sub run_with_timeout {
    my ($timeout, $code, @param) = @_;

    die "got timeout\n" if $timeout <= 0;

    my $prev_alarm;

    my $sigcount = 0;

    my $res;

    eval {
	local $SIG{ALRM} = sub { $sigcount++; die "got timeout\n"; };
	local $SIG{PIPE} = sub { $sigcount++; die "broken pipe\n" };
	local $SIG{__DIE__};   # see SA bug 4631

	$prev_alarm = alarm ($timeout);

	$res = &$code (@param);

	alarm 0; # avoid race conditions
    };

    my $err = $@;
    
    alarm ($prev_alarm) if defined ($prev_alarm);

    die "unknown error" if $sigcount && !$err; # seems to happen sometimes

    die $err if $err;

    return $res;
}

sub cluster_lock_storage {
    my ($storeid, $shared, $timeout, $func, @param) = @_;

    my $ccfg = load_cluster_config();

    my $masterip = $ccfg && $ccfg->{master} ? $ccfg->{master}->{ip} : undef;
   
    if (!$shared || !$masterip) {
	return &$func (@param);
    }

    $timeout = 60 if !$timeout;

    my $cmdpipe;
    my $got_lock;

    my $ret_value;

    eval {

	my $cmd;

	if ($masterip eq $ccfg->{local}->{ip}) {
	    $cmd = ['/usr/sbin/pvesm', 'lock', $storeid, $timeout];
	} else {
	    $cmd = ['/usr/bin/ssh', '-o', 'BatchMode=yes', $masterip,
		    '/usr/sbin/pvesm', 'lock', $storeid, $timeout];
	}

	$cmdpipe = fork_command_pipe ($cmd);

	my $helo;
	my $reader = $cmdpipe->{reader};

	eval { run_with_timeout ($timeout + 5, sub { $helo = <$reader>; }); };
	die "can't lock storage '$storeid' - $@" if $@;
	die "can't lock storage '$storeid' - no reply\n" if !$helo;
	if ($helo !~ m/^lock $storeid: (.*)$/) {
	    die "got strange reply from lock ('$helo')\n";
	}
	my $res = $1;
	die "can't aquire lock - $res\n" if $res ne 'success';

	$got_lock = 1;

	$ret_value = &$func (@param);

    };
    my $err = $@;
    
    if ($cmdpipe) {

	if ($got_lock) {
	    my $writer = $cmdpipe->{writer};

	    eval { 
		run_with_timeout (60, sub {
		    print $writer "release\n";
		    $writer->flush();
		}); 
	    };
	    warn $@ if $@;
	}
	finish_command_pipe ($cmdpipe);
    }
    
    die $err if $err;

    return $ret_value;
}


sub storage_config {
    my ($cfg, $storeid, $noerr) = @_;

    my $scfg = $cfg->{ids}->{$storeid};

    die "storage '$storeid' does not exists\n" if (!$noerr && !$scfg);

    return $scfg;
}

sub storage_check_enabled {
    my ($cfg, $storeid) = @_;

    my $scfg = storage_config ($cfg, $storeid);

    die "storage '$storeid' is disabled\n" if $scfg->{disable};

    return $scfg;
}

sub storage_ids {
    my ($cfg) = @_;

    my $ids = $cfg->{ids};

    my @sa = sort {$ids->{$a}->{priority} <=> $ids->{$b}->{priority}} keys %$ids;

    return @sa;
}

sub __assert_if_modified {
    my ($cfg, $digest) = @_;

    if ($digest && ($cfg->{digest} ne $digest)) {
	die "detected modified storage configuration - try again\n";
    }
}

sub sprint_config_line {
    my ($k, $v) = @_;

    my $ct = $confvars->{$k};

    if ($ct eq 'bool') {
	return $v ? "\t$k\n" : '';
    } elsif ($ct eq 'content') {
	my $clist = join (',', keys %$v);
	if ($clist) {
	    return "\t$k $clist\n";
	} else {
	    return "\t$k none\n";
	}
    } else {
	return "\t$k $v\n";
    }
}

sub write_config {
    my ($cfg) = @_;

    my $tmpfn = "$storageconfig.tmp$$";
    my $fh = IO::File->new ($tmpfn, 'w') ||
	die "unable to write file '$tmpfn' - $!";

    my $ids = $cfg->{ids};

    eval {
	my $maxpri = 0;
	foreach my $storeid (keys %$ids) {
	    my $pri = $ids->{$storeid}->{priority}; 
	    $maxpri = $pri if $pri && $pri > $maxpri;
	}
	foreach my $storeid (keys %$ids) {
	    if (!defined ($ids->{$storeid}->{priority})) {
		$ids->{$storeid}->{priority} = ++$maxpri;
	    } 
	}

	foreach my $storeid (sort {$ids->{$a}->{priority} <=> $ids->{$b}->{priority}} keys %$ids) {
	    my $scfg = $ids->{$storeid};
	    my $type = $scfg->{type};
	    my $def = $default_config->{$type};

	    die "unknown storage type '$type'\n" if !$def;

	    my $data = "$type: $storeid\n";

	    $data .= "\tdisable\n" if $scfg->{disable};

	    my $done_hash = { disable => 1};
	    foreach my $k (@{$required_config->{$type}}) {
		$done_hash->{$k} = 1;
		my $v =  $ids->{$storeid}->{$k};
		die "storage '$storeid' - missing value for required option '$k'\n"
		    if !defined ($v);
		$data .= sprint_config_line ($k, $v);
	    }

	    foreach my $k (keys %$def) {
		next if defined ($done_hash->{$k});
		if (defined (my $v = $ids->{$storeid}->{$k})) {
		    $data .= sprint_config_line ($k, $v);
		}
	    }

	    print $fh "$data\n";
	}
    };

    my $err = $@;

    close ($fh);

    if ($err) {
	unlink $tmpfn;
	die $err;
    } else {
	rename $tmpfn, $storageconfig;
    }
}

sub get_image_dir {
    my ($cfg, $storeid, $vmid) = @_;

    my $path = $cfg->{ids}->{$storeid}->{path};
    return $vmid ? "$path/images/$vmid" : "$path/images";
}

sub get_iso_dir {
    my ($cfg, $storeid) = @_;

    my $isodir =  $cfg->{ids}->{$storeid}->{path};
    $isodir .= '/template/iso' if $storeid eq 'local';

    return $isodir;
}

sub get_vztmpl_dir {
    my ($cfg, $storeid) = @_;

    my $tmpldir =  $cfg->{ids}->{$storeid}->{path};
    $tmpldir .= '/template/cache' if $storeid eq 'local';

    return $tmpldir;
}

# iscsi utility functions

sub iscsi_session_list {

    check_iscsi_support ();

    my $cmd = [$ISCSIADM, '--mode', 'session'];

    my $res = {};

    run_command ($cmd, outfunc => sub {
	my $line = shift;

	if ($line =~ m/^tcp:\s+\[(\S+)\]\s+\S+\s+(\S+)\s*$/) {
	    my ($session, $target) = ($1, $2);
	    $res->{$target} = $session;
	}
    });

    return $res;
}

sub iscsi_discovery {
    my ($portal) = @_;

    check_iscsi_support ();

    my $cmd = [$ISCSIADM, '--mode', 'discovery', '--type', 'sendtargets', 
	       '--portal', $portal];

    my $res = {};
    run_command ($cmd, outfunc => sub {
	my $line = shift;

	if ($line =~ m/^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d+)\,\S+\s+(\S+)\s*$/) {
	    my $portal = $1;
	    my $target = $2;
	    $res->{$target} = $portal;
	}
    });

    return $res;
}

sub iscsi_login {
    my ($target, $portal) = @_;

    check_iscsi_support ();

    iscsi_discovery ($portal);

    my $cmd = [$ISCSIADM, '--mode', 'node', '--portal', $portal, 
	       '--targetname',  $target, '--login'];
    run_command ($cmd);
}

sub iscsi_logout {
    my ($target, $portal) = @_;

    check_iscsi_support ();

    my $cmd = [$ISCSIADM, '--mode', 'node', '--portal', $portal, '--targetname', 
	       $target, '--logout'];
    run_command ($cmd);
}

my $rescan_filename = "/var/run/pve-iscsi-rescan.lock";

sub iscsi_session_rescan {
    my $session = shift;

    check_iscsi_support ();

    my $rstat = stat ($rescan_filename);

    if (!$rstat) {
	if (my $fh = IO::File->new ($rescan_filename, "a")) {
	    utime undef, undef, $fh;
	    close ($fh);
	}
    } else {
	my $atime = $rstat->atime;
	my $tdiff = time() - $atime;
	# avoid frequent rescans
	return if !($tdiff < 0 || $tdiff > 10);
	utime undef, undef, $rescan_filename;
    }

    my $cmd = [$ISCSIADM, '--mode', 'session', '-r', $session, '-R'];
    eval { run_command ($cmd, outfunc => sub {}); };
    warn $@ if $@;
}

sub iscsi_device_list {

    my $res = {};

    my $dirname = '/sys/class/iscsi_session';

    dir_glob_foreach ($dirname, 'session(\d+)', sub {
	my ($ent, $session) = @_;

	my $target = file_read_firstline ("$dirname/$ent/targetname");
	return if !$target;

	my (undef, $host) = dir_glob_regex ("$dirname/$ent/device", 'target(\d+):.*');
	return if !defined($host);

	dir_glob_foreach ("/sys/bus/scsi/devices", "$host:" . '(\d+):(\d+):(\d+)', sub {
	    my ($tmp, $channel, $id, $lun) = @_;

	    my $type = file_read_firstline ("/sys/bus/scsi/devices/$tmp/type");
	    return if !defined($type) || $type ne '0'; # list disks only

	    my $bdev;
	    if (-d "/sys/bus/scsi/devices/$tmp/block") { # newer kernels
		(undef, $bdev) = dir_glob_regex ("/sys/bus/scsi/devices/$tmp/block/", '([A-Za-z]\S*)');
	    } else {
		(undef, $bdev) = dir_glob_regex ("/sys/bus/scsi/devices/$tmp", 'block:(\S+)');
	    }
	    return if !$bdev;

	    my $blockdev = find_dev_by_id ("/dev/$bdev");
	    return if !$blockdev;

	    my $size = file_read_firstline ("/sys/block/$bdev/size");
	    return if !$size;

	    my $volid = "$channel.$id.$lun.$blockdev";

	    $res->{$target}->{$volid} = {
		'format' => 'raw', 
		'size' => int($size / 2), 
		'vmid' => 0, # not assigned to any vm
		'channel' => $channel,
		'id' => $id,
		'lun' => $lun,
	    };

	    #print "TEST: $target $session $host,$bus,$tg,$lun $blockdev\n"; 
	});

    });

    return $res;
}

# library implementation

sub parse_storage_id {
    my ($storeid, $noerr) = @_;

    if ($storeid !~ m/^[a-z][a-z0-9\-\_\.]*[a-z0-9]$/i) {
	return undef if $noerr;
	die "storage ID '$storeid' contains illegal characters\n";
    }
    return $storeid;
}

sub parse_lvm_name {
    my ($name, $noerr) = @_;

    if ($name !~ m/^[a-z][a-z0-9\-\_\.]*[a-z0-9]$/i) {
	return undef if $noerr;
	die "lvm name '$name' contains illegal characters\n";
    }

    return $name;
}

sub parse_vmid {
    my $vmid = shift;

    die "VMID '$vmid' contains illegal characters\n" if $vmid !~ m/^\d+$/;

    return int($vmid);
}

sub parse_volume_id {
    my ($volid, $noerr) = @_;

    if ($volid =~ m/^([a-z][a-z0-9\-\_\.]*[a-z0-9]):(.+)$/i) {
	return wantarray ? ($1, $2) : $1;
    }
    return undef if $noerr;
    die "unable to parse volume ID '$volid'\n";
}

sub parse_name_dir {
    my $name = shift;

    if ($name =~ m!^([^/\s]+\.(raw|qcow2|vmdk))$!) {
	return ($1, $2);
    }

    die "unable to parse volume filename '$name'\n";
}

sub parse_volname_dir {
    my $volname = shift;

    if ($volname =~ m!^(\d+)/(\S+)$!) {
	my ($vmid, $name) = ($1, $2);
	parse_name_dir ($name);
	return ('image', $name, $vmid);
    } elsif ($volname =~ m!^iso/([^/]+\.[Ii][Ss][Oo])$!) {
	return ('iso', $1);
    } elsif ($volname =~ m!^vztmpl/([^/]+\.tar\.gz)$!) {
	return ('vztmpl', $1);
    }
    die "unable to parse directory volume name '$volname'\n";
}

sub parse_volname_lvm {
    my $volname = shift;

    parse_lvm_name ($volname);

    if ($volname =~ m/^(vm-(\d+)-\S+)$/) {
	return ($1, $2);
    }

    die "unable to parse lvm volume name '$volname'\n";    
}

sub parse_volname_iscsi {
    my $volname = shift;

    if ($volname =~ m!^\d+\.\d+\.\d+\.(\S+)$!) {
	my $byid = $1;
	return $byid;
    }

    die "unable to parse iscsi volume name '$volname'\n";
}

# try to map a filesystem path to a volume identifier
sub path_to_volume_id {
    my ($cfg, $path) = @_;

    my $ids = $cfg->{ids};

    my ($sid, $volname) = parse_volume_id ($path, 1);
    if ($sid) {
	if ($ids->{$sid} && (my $type = $ids->{$sid}->{type})) {
	    if ($type eq 'dir' || $type eq 'nfs') {
		my ($vtype, $name, $vmid) = parse_volname_dir ($volname);
		return ($vtype, $path);
	    }
	}
	return ('');
    }

    $path = abs_path ($path);

    foreach my $sid (keys %$ids) {
	my $type = $ids->{$sid}->{type};
	next if !($type eq 'dir' || $type eq 'nfs');
	
	my $imagedir = $ids->{$sid}->{path} . "/images";
	my $isodir = get_iso_dir ($cfg, $sid);
	my $tmpldir = get_vztmpl_dir ($cfg, $sid);

	if ($path =~ m!^$imagedir/(\d+)/([^/\s]+)$!) {
	    my $vmid = $1;
	    my $name = $2;
	    return ('image', "$sid:$vmid/$name");
	} elsif ($path =~ m!^$isodir/([^/]+\.[Ii][Ss][Oo])$!) {
	    my $name = $1;
	    return ('iso', "$sid:iso/$name");	
	} elsif ($path =~ m!^$tmpldir/([^/]+\.tar\.gz)$!) {
	    my $name = $1;
	    return ('vztmpl', "$sid:vztmpl/$name");
	}
    }

    # can't map path to volume id
    return ('');
}

sub path {
    my ($cfg, $volid) = @_;

    my ($storeid, $volname) = parse_volume_id ($volid);

    my $scfg = storage_config ($cfg, $storeid);

    my $path;
    my $owner;

    if ($scfg->{type} eq 'dir' || $scfg->{type} eq 'nfs') {
	my ($vtype, $name, $vmid) = parse_volname_dir ($volname);
	$owner = $vmid;

	my $imagedir = get_image_dir ($cfg, $storeid, $vmid);
	my $isodir = get_iso_dir ($cfg, $storeid);
	my $tmpldir = get_vztmpl_dir ($cfg, $storeid);

	if ($vtype eq 'image') {
	    $path = "$imagedir/$name";
	} elsif ($vtype eq 'iso') {
	    $path = "$isodir/$name";
	} elsif ($vtype eq 'vztmpl') {
	    $path = "$tmpldir/$name";
	} else {
	    die "should not be reached";
	}

    } elsif ($scfg->{type} eq 'lvm') {

	my $vg = $scfg->{vgname};

	my ($name, $vmid) = parse_volname_lvm ($volname);
	$owner = $vmid;

	$path = "/dev/$vg/$name";

    } elsif ($scfg->{type} eq 'iscsi') {
	my $byid = parse_volname_iscsi ($volname);
	$path = "/dev/disk/by-id/$byid";
    } else {
	die "unknown storage type '$scfg->{type}'";
    }

    return wantarray ? ($path, $owner) : $path;
}

sub storage_migrate {
    my ($cfg, $volid, $target_host, $target_storeid, $target_volname) = @_;

    my ($storeid, $volname) = parse_volume_id ($volid);
    $target_volname = $volname if !$target_volname;

    my $scfg = storage_config ($cfg, $storeid);

    # no need to migrate shared content
    return if $storeid eq $target_storeid && $scfg->{shared};

    my $tcfg = storage_config ($cfg, $target_storeid);

    my $target_volid = "${target_storeid}:${target_volname}";

    my $errstr = "unable to migrate '$volid' to '${target_volid}' on host '$target_host'";

    # blowfish is a fast block cipher, much faster then 3des
    my $sshoptions = "-c blowfish -o 'BatchMode=yes'";
    my $ssh = "/usr/bin/ssh $sshoptions";

    local $ENV{RSYNC_RSH} = $ssh;

    if ($scfg->{type} eq 'dir' || $scfg->{type} eq 'nfs') {
	if ($tcfg->{type} eq 'dir' || $tcfg->{type} eq 'nfs') {

	    my $src = path ($cfg, $volid);
	    my $dst = path ($cfg, $target_volid);

	    my $dirname = dirname ($dst);

	    if ($tcfg->{shared}) { # we can do a local copy
		
		run_command (['/bin/mkdir', '-p', $dirname]);

		run_command (['/bin/cp', $src, $dst]);

	    } else {

		run_command (['/usr/bin/ssh', "root\@${target_host}", 
			      '/bin/mkdir', '-p', $dirname]);

		my $cmd = ['/usr/bin/rsync', '--progress', '--inplace', '--no-whole-file', 
			   $src, "root\@${target_host}:$dst"];

		my $percent = -1;

		run_command ($cmd, outfunc => sub {
		    my $line = shift;

		    if ($line =~ m/^\s*(\d+\s+(\d+)%\s.*)$/) {
			if ($2 > $percent) {
			    $percent = $2;
			    print "rsync status: $1\n";
			    *STDOUT->flush();
			}
		    } else {
			print "$line\n";
			*STDOUT->flush();
		    }
		});
	    }


	} else {

	    die "$errstr - target type '$tcfg->{type}' not implemented\n";
	}

    } else {
	die "$errstr - source type '$scfg->{type}' not implemented\n";
    }
}

sub vdisk_alloc {
    my ($cfg, $storeid, $vmid, $fmt, $name, $size) = @_;

    die "no storage id specified\n" if !$storeid;

    parse_storage_id ($storeid);

    my $scfg = storage_config ($cfg, $storeid);

    die "no VMID specified\n" if !$vmid;

    $vmid = parse_vmid ($vmid);

    my $defformat = storage_default_format ($cfg, $storeid);

    $fmt = $defformat if !$fmt;

    activate_storage ($cfg, $storeid);

    # lock shared storage
    return cluster_lock_storage ($storeid, $scfg->{shared}, undef, sub {

	if ($scfg->{type} eq 'dir' || $scfg->{type} eq 'nfs') {

	    my $imagedir = get_image_dir ($cfg, $storeid, $vmid);

	    mkpath $imagedir;

	    if (!$name) {

		for (my $i = 1; $i < 100; $i++) {
		    my @gr = <$imagedir/vm-$vmid-disk-$i.*>;
		    if (!scalar(@gr)) {
			$name = "vm-$vmid-disk-$i.$fmt";
			last;
		    }
		}
	    }

	    die "unable to allocate an image name for VM $vmid in storage '$storeid'\n"
		if !$name;

	    my (undef, $tmpfmt) = parse_name_dir ($name);

	    die "illegal name '$name' - wrong extension for format ('$tmpfmt != '$fmt')\n" 
		if $tmpfmt ne $fmt;

	    my $path = "$imagedir/$name";

	    die "disk image '$path' already exists\n" if -f $path;

	    if (system ("/usr/bin/qemu-img create -f $fmt '$path' ${size}K") != 0) {
		my $msg = "unable to create image - $!\n";
		unlink $path;
		die $msg;
	    }

	    return "$storeid:$vmid/$name";

	} elsif ($scfg->{type} eq 'lvm') {
	
	    die "unsupported format '$fmt'" if $fmt ne 'raw';

	    die "illegal name '$name' - sould be 'vm-$vmid-*'\n" 
		if  $name && $name !~ m/^vm-$vmid-/;

	    my $vgs = lvm_vgs ();

	    my $vg = $scfg->{vgname};

	    die "no such volume gruoup '$vg'\n" if !defined ($vgs->{$vg});

	    my $free = int ($vgs->{$vg}->{free});

	    die "not enough free space ($free < $size)\n" if $free < $size;

	    if (!$name) {
		my $lvs = lvm_lvs ($vg);

		for (my $i = 1; $i < 100; $i++) {
		    my $tn = "vm-$vmid-disk-$i";
		    if (!defined ($lvs->{$vg}->{$tn})) {
			$name = $tn;
			last;
		    }
		}
	    }

	    die "unable to allocate an image name for VM $vmid in storage '$storeid'\n"
		if !$name;

	    my $cmd = ['/sbin/lvcreate', '--addtag', "pve-vm-$vmid", '--size', "${size}k", '--name', $name, $vg];

	    run_command ($cmd);

	    return "$storeid:$name";

	} elsif ($scfg->{type} eq 'iscsi') {
	    die "can't allocate space in iscsi storage\n";
	} else {
	    die "unknown storage type '$scfg->{type}'";
	}
    });
}

sub vdisk_free {
    my ($cfg, $volid) = @_;

    my ($storeid, $volname) = PVE::Storage::parse_volume_id ($volid);

    my $scfg = storage_config ($cfg, $storeid);

    activate_storage ($cfg, $storeid);

    # lock shared storage
    cluster_lock_storage ($storeid, $scfg->{shared}, undef, sub {

	if ($scfg->{type} eq 'dir' || $scfg->{type} eq 'nfs') {
	    my $path = path ($cfg, $volid); 

	    if (! -f $path) {
		warn "disk image '$path' does not exists\n";
	    } else {
		unlink $path;
	    }
	} elsif ($scfg->{type} eq 'lvm') {

	    my $vg = $scfg->{vgname};

	    my $cmd = ['/sbin/lvremove', '-f', "$vg/$volname"];

	    run_command ($cmd);
	} elsif ($scfg->{type} eq 'iscsi') {
	    die "can't free space in iscsi storage\n";
	} else {
	    die "unknown storage type '$scfg->{type}'";
	}
    });
}

# lvm utility functions

sub lvm_pv_info {
    my ($device) = @_;

    die "no device specified" if !$device;

    my $has_label = 0;

    my $cmd = ['/usr/bin/file', '-L', '-s', $device];
    run_command ($cmd, outfunc => sub {
	my $line = shift;
	$has_label = 1 if $line =~ m/LVM2/;
    });

    return undef if !$has_label;

    $cmd = ['/sbin/pvs', '--separator', ':', '--noheadings', '--units', 'k',
	    '--unbuffered', '--nosuffix', '--options',
	    'pv_name,pv_size,vg_name,pv_uuid', $device];

    my $pvinfo;
    run_command ($cmd, outfunc => sub {
	my $line = shift;

	$line =~ s/^\s+//;
	$line =~ s/\s+$//;

	my ($pvname, $size, $vgname, $uuid) = split (':', $line);

	die "found multiple pvs entries for device '$device'\n" 
	    if $pvinfo;

	$pvinfo = {
	    pvname => $pvname,
	    size => $size,
	    vgname => $vgname,
	    uuid => $uuid,
	};
    });

    return $pvinfo;
}

sub clear_first_sector {
    my ($dev) = shift;

    if (my $fh = IO::File->new ($dev, "w")) {
	my $buf = 0 x 512;
	syswrite $fh, $buf;
	$fh->close();	
    }
}


sub lvm_create_volume_group {
    my ($device, $vgname, $shared) = @_;
    
    my $res = lvm_pv_info ($device);
    
    if ($res->{vgname}) {
	return if $res->{vgname} eq $vgname; # already created
	die "device '$device' is already used by volume group '$res->{vgname}'\n";
    }

    clear_first_sector ($device); # else pvcreate fails

    # we use --metadatasize 250k, which reseults in "pe_start = 512"
    # so pe_start is aligned on a 128k boundary (advantage for SSDs)
    my $cmd = ['/sbin/pvcreate', '--metadatasize', '250k', $device];

    run_command ($cmd);

    $cmd = ['/sbin/vgcreate', $vgname, $device];
    # push @$cmd, '-c', 'y' if $shared; # we do not use this yet

    run_command ($cmd);
}

sub lvm_vgs {

    my $cmd = ['/sbin/vgs', '--separator', ':', '--noheadings', '--units', 'k',
	       '--unbuffered', '--nosuffix', '--options',
	       'vg_name,vg_size,vg_free'];

    my $vgs = {};
    run_command ($cmd, outfunc => sub {
	my $line = shift;

	$line =~ s/^\s+//;
	$line =~ s/\s+$//;

	my ($name, $size, $free) = split (':', $line);

	$vgs->{$name} = { size => int ($size), free => int ($free) };
    });

    return $vgs;
}

sub lvm_lvs {
    my ($vgname) = @_;

    my $cmd = ['/sbin/lvs', '--separator', ':', '--noheadings', '--units', 'k',
	       '--unbuffered', '--nosuffix', '--options',
	       'vg_name,lv_name,lv_size,uuid,tags'];

    push @$cmd, $vgname if $vgname;

    my $lvs = {};
    run_command ($cmd, outfunc => sub {
	my $line = shift;

	$line =~ s/^\s+//;
	$line =~ s/\s+$//;

	my ($vg, $name, $size, $uuid, $tags) = split (':', $line);

	return if $name !~ m/^vm-(\d+)-/;
	my $nid = $1;

	my $owner;
	foreach my $tag (split (/,/, $tags)) {
	    if ($tag =~ m/^pve-vm-(\d+)$/) {
		$owner = $1;
		last;
	    }
	}
	
	if ($owner) {
	    if ($owner ne $nid) {
		warn "owner mismatch name = $name, owner = $owner\n";
	    }
   
	    $lvs->{$vg}->{$name} = { format => 'raw', size => $size, 
				     uuid => $uuid,  tags => $tags, 
				     vmid => $owner };
	}
    });

    return $lvs;
}

#install iso or openvz template ($tt = <iso|vztmpl>)
# we simply overwrite when file already exists
sub install_template {
    my ($cfg, $storeid, $tt, $srcfile, $destfile) = @_;

    my $scfg = storage_config ($cfg, $storeid);

    my $type = $scfg->{type};

    die "invalid storage type '$type'" if !($type eq 'dir' || $type eq 'nfs');

    my $path;

    if ($tt eq 'iso') {
	die "file '$destfile' has no '.iso' extension\n" 
	    if $destfile !~ m![^/]+\.[Ii][Ss][Oo]$!;
	die "storage '$storeid' does not support 'iso' content\n" 
	    if !$scfg->{content}->{iso};
	$path = get_iso_dir ($cfg, $storeid);
    } elsif ($tt eq 'vztmpl') {
	die "file '$destfile' has no '.tar.gz' extension\n"
	    if $destfile !~ m![^/]+\.tar\.gz$!;
	die "storage '$storeid' does not support 'vztmpl' content\n" 
	    if !$scfg->{content}->{vztmpl};
	$path = get_vztmpl_dir ($cfg, $storeid);
    } else {
	die "unknown template type '$tt'";
    }

    activate_storage ($cfg, $storeid);

    my $dest = "$path/$destfile";

    my $cmd = ['cp', $srcfile, $dest];

    eval { run_command ($cmd); };
    my $err = $@;

    if ($err) {
	unlink $dest;
	die $err;
    }
}

#list iso or openvz template ($tt = <iso|vztmpl|backup>)
sub template_list {
    my ($cfg, $storeid, $tt) = @_;

    die "unknown template type" if !($tt eq 'iso' || $tt eq 'vztmpl' || $tt eq 'backup'); 

    my $ids = $cfg->{ids};

    storage_check_enabled ($cfg, $storeid) if ($storeid);

    my $res = {};

    # query the storage

    foreach my $sid (keys %$ids) {
	next if $storeid && $storeid ne $sid;

	my $scfg = $ids->{$sid};
	my $type = $scfg->{type};

	next if $scfg->{disable};

	next if $tt eq 'iso' && !$scfg->{content}->{iso};
	next if $tt eq 'vztmpl' && !$scfg->{content}->{vztmpl};
	next if $tt eq 'backup' && !$scfg->{content}->{backup};

	activate_storage ($cfg, $sid);

	if ($type eq 'dir' || $type eq 'nfs') {

	    my $path;
	    if ($tt eq 'iso') {
		$path = get_iso_dir ($cfg, $sid);
	    } elsif ($tt eq 'vztmpl') {
		$path = get_vztmpl_dir ($cfg, $sid);
	    } elsif ($tt eq 'backup') {
		$path = $scfg->{path};
	    } else {
		die "unknown template type";
	    }

	    foreach my $fn (<$path/*>) {

		my $info;

		if ($tt eq 'iso') {
		    next if $fn !~ m!/([^/]+\.[Ii][Ss][Oo])$!;

		    $info = { volname => "iso/$1", format => 'iso' };

		} elsif ($tt eq 'vztmpl') {
		    next if $fn !~ m!/([^/]+\.tar\.gz)$!;

		    $info = { volname => "vztmpl/$1", format => 'tgz' };

		} elsif ($tt eq 'backup') {
		    next if $fn !~ m!/([^/]+\.(tar|tgz))$!;
		    
		    $info = { volname => "backup/$1", format => $2 };
		}

		$info->{size} = -s $fn;

		push @{$res->{$sid}}, $info;
	    }

	}

	@{$res->{$sid}} = sort {lc($a->{volname}) cmp lc ($b->{volname}) } @{$res->{$sid}} if $res->{$sid};
    }

    return $res;
}

sub file_size_info {
    my ($filename) = @_;

    my $cmd = ['/usr/bin/qemu-img', 'info', $filename];

    my $format;
    my $size = 0;
    my $used = 0;

    eval {
	run_command ($cmd, outfunc => sub {
	    my $line = shift;

	    if ($line =~ m/^file format:\s+(\S+)\s*$/) {
		$format = $1;
	    } elsif ($line =~ m/^virtual size:\s\S+\s+\((\d+)\s+bytes\)$/) {
		$size = int ($1/1024);
	    } elsif ($line =~ m/^disk size:\s+(\d+(.\d+)?)([KMGT])\s*$/) {
		$used = $1;
		my $u = $3;

		$used *= 1 if $u eq 'K';
		$used *= 1024 if $u eq 'M';
		$used *= (1024*1024) if $u eq 'G';
		$used *= (1024*1024*1024) if $u eq 'T';
	    }
	});
    };

    return wantarray ? ($size, $format, $used) : $size;
}

sub vdisk_list {
    my ($cfg, $storeid, $vmid, $vollist) = @_;

    my $ids = $cfg->{ids};

    storage_check_enabled ($cfg, $storeid) if ($storeid);

    my $res = {};

    # prepare/activate/refresh all storages

    my $stypes = {};

    my $storage_list = [];
    if ($vollist) {
	foreach my $volid (@$vollist) {
	    my ($sid, undef) = parse_volume_id ($volid);
	    next if !defined ($ids->{$sid});
	    next if $ids->{$sid}->{disable};
	    push @$storage_list, $sid;
	    $stypes->{$ids->{$sid}->{type}} = 1;
	}
    } else {
	foreach my $sid (keys %$ids) {
	    next if $storeid && $storeid ne $sid;
	    next if $ids->{$sid}->{disable};

	    push @$storage_list, $sid;
	    $stypes->{$ids->{$sid}->{type}} = 1;
	}
    }

    activate_storage_list ($cfg, $storage_list);

    my $lvs = $stypes->{lvm} ? lvm_lvs () : {};

    my $iscsi_devices = iscsi_device_list() if $stypes->{iscsi};

    # query the storage

    foreach my $sid (keys %$ids) {
	if ($storeid) {
	    next if $storeid ne $sid;
	    next if $ids->{$storeid}->{disable};
	}
	my $scfg = $ids->{$sid};
	my $type = $scfg->{type};

	if ($type eq 'dir' || $type eq 'nfs') {

	    my $path = $scfg->{path};

	    my $fmts = join ('|', keys %{$default_config->{$type}->{format}->[0]}); 

	    foreach my $fn (<$path/images/[0-9][0-9]*/*>) {

		next if $fn !~ m!/images/(\d+)/([^/]+\.($fmts))$!;

		my $owner = $1;
		my $name = $2;
		my $volname = "$owner/$name";

		if ($vollist) {
		    my $found = grep { $_ eq "$sid:$volname" } @$vollist;
		    next if !$found;
		} else {
		    next if defined ($vmid) && ($owner ne $vmid);
		}

		my ($size, $format, $used) = file_size_info ($fn);

		if ($format && $size) {
		    push @{$res->{$sid}}, { 
			volname => $volname, format => $format,
			size => $size, vmid => $owner, used => $used };
		}

	    }

	} elsif ($type eq 'lvm') {

	    my $vgname = $scfg->{vgname};

	    if (my $dat = $lvs->{$vgname}) {

		foreach my $volname (keys %$dat) {

		    my $owner = $dat->{$volname}->{vmid};

		    if ($vollist) {
			my $found = grep { $_ eq "$sid:$volname" } @$vollist;
			next if !$found;
		    } else {
			next if defined ($vmid) && ($owner ne $vmid);
		    }

		    my $info = $dat->{$volname};
		    $info->{volname} = $volname;

		    push @{$res->{$sid}}, $info;
		}
	    }

	} elsif ($type eq 'iscsi') {

	    # we have no owner for iscsi devices

	    my $target = $scfg->{target};

	    if (my $dat = $iscsi_devices->{$target}) {

		foreach my $volname (keys %$dat) {

		    if ($vollist) {
			my $found = grep { $_ eq "$sid:$volname" } @$vollist;
			next if !$found;
		    } else {
			next if !($storeid && ($storeid eq $sid));
		    }

		    my $info = $dat->{$volname};
		    $info->{volname} = $volname;

		    push @{$res->{$sid}}, $info;
		}
	    }

	} else {
	    die "implement me";
	}

	@{$res->{$sid}} = sort {lc($a->{volname}) cmp lc ($b->{volname}) } @{$res->{$sid}} if $res->{$sid};
    }

    return $res;
}

sub nfs_is_mounted {
    my ($server, $export, $mountpoint, $mountdata) = @_;

    my $source = "$server:$export";

    $mountdata = read_proc_mounts() if !$mountdata;

    if ($mountdata =~ m/^$source\s$mountpoint\snfs/m) {
	return $mountpoint;
    } 

    return undef;
}

sub nfs_mount {
    my ($server, $export, $mountpoint, $options) = @_;

    my $source = "$server:$export";

    my $cmd = ['/bin/mount', '-t', 'nfs', $source, $mountpoint];
    if ($options) {
	push @$cmd, '-o', $options;
    }

    run_command ($cmd);
}

sub uevent_seqnum {

    my $filename = "/sys/kernel/uevent_seqnum";

    my $seqnum = 0;
    if (my $fh = IO::File->new ($filename, "r")) {
	my $line = <$fh>;
	if ($line =~ m/^(\d+)$/) {
	    $seqnum = int ($1);
	}
	close ($fh);
    }
    return $seqnum;
}

sub __activate_storage_full {
    my ($cfg, $storeid, $session) = @_;

    my $scfg = storage_check_enabled ($cfg, $storeid);

    return if $session->{activated}->{$storeid};

    if (!$session->{mountdata}) {
	$session->{mountdata} = read_proc_mounts();
    }

    if (!$session->{uevent_seqnum}) {
	$session->{uevent_seqnum} = uevent_seqnum ();
    }

    my $mountdata = $session->{mountdata};

    my $type = $scfg->{type};

    if ($type eq 'dir' || $type eq 'nfs') {

	my $path = $scfg->{path};
	mkpath $path;

	die "unable to activate storage '$storeid' - " .
	    "directory '$path' does not exist\n" if ! -d $path;

	if ($type eq 'nfs') {
	    my $server = $scfg->{server};
	    my $export = $scfg->{export};

	    if (!nfs_is_mounted ($server, $export, $path, $mountdata)) {
		nfs_mount ($server, $export, $path, $scfg->{options});
	    }
	}

	my $imagedir = get_image_dir ($cfg, $storeid);
	my $isodir = get_iso_dir ($cfg, $storeid);
	my $tmpldir = get_vztmpl_dir ($cfg, $storeid);

	mkpath $imagedir if $scfg->{content}->{images} &&
	    $imagedir ne $path;
	mkpath $isodir if $scfg->{content}->{iso} &&
	    $isodir ne $path;
	mkpath $tmpldir if $scfg->{content}->{vztmpl} &&
	    $tmpldir ne $path;

    } elsif ($type eq 'lvm') {

	if ($scfg->{base}) {
	    my ($baseid, undef) = parse_volume_id ($scfg->{base});
	    __activate_storage_full ($cfg, $baseid, $session);
	}

	# In LVM2, vgscans take place automatically;
	# this is just to be sure
	my $res = lvm_vgs (); 
	if (!$res->{$scfg->{vgname}}) {
	    my $cmd = ['/sbin/vgscan', '--ignorelockingfailure', '--mknodes'];
	    eval { run_command ($cmd, outfunc => sub {}); };
	    warn $@ if $@;
	}

	my $cmd = ['/sbin/vgchange', '-aly', $scfg->{vgname}];
	run_command ($cmd, outfunc => sub {});

    } elsif ($type eq 'iscsi') {

	return if !check_iscsi_support(1);

	$session->{iscsi_sessions} = iscsi_session_list()
	    if !$session->{iscsi_sessions};

	my $iscsi_sess = $session->{iscsi_sessions}->{$scfg->{target}};
	if (!defined ($iscsi_sess)) {
	    eval { iscsi_login ($scfg->{target}, $scfg->{portal}); };
	    warn $@ if $@;
	} else {
	    # make sure we get all devices
	    iscsi_session_rescan ($iscsi_sess);
	}

    } else {
	die "implement me";
    }

    my $newseq = uevent_seqnum ();

    # only call udevsettle if there are events
    if ($newseq > $session->{uevent_seqnum}) {
	my $timeout = 30;
	system ("$UDEVADM settle --timeout=$timeout"); # ignore errors
	$session->{uevent_seqnum} = $newseq;
    }

    $session->{activated}->{$storeid} = 1;
}

sub activate_storage_list {
    my ($cfg, $storeid_list) = @_;

    my $session = {};

    foreach my $storeid (@$storeid_list) {
	__activate_storage_full ($cfg, $storeid, $session);
    }
}

sub activate_storage {
    my ($cfg, $storeid) = @_;

    my $session = {};

    __activate_storage_full ($cfg, $storeid, $session);
}

sub activate_volumes {
    my ($cfg, $vollist) = @_;

    my $storagehash = {};
    foreach my $volid (@$vollist) {
	my ($storeid, undef) = parse_volume_id ($volid);
	$storagehash->{$storeid} = 1;
    }

    activate_storage_list ($cfg, [keys %$storagehash]);

    foreach my $volid (@$vollist) {
	my ($storeid, $volname) = parse_volume_id ($volid);

	my $scfg = storage_config ($cfg, $storeid);

	my $path = path ($cfg, $volid);

	if ($scfg->{type} eq 'lvm') {
	    my $cmd = ['/sbin/lvchange', '-aly', $path];
	    eval { run_command ($cmd); };
	    warn $@ if $@;
	}

	# check is volume exists
	if ($scfg->{type} eq 'dir' || $scfg->{type} eq 'nfs') {
	    die "volume '$volid' does not exist\n" if ! -f $path;
	} else {
	    die "volume '$volid' does not exist\n" if ! -b $path;
	}
    }
}

sub deactivate_volumes {
    my ($cfg, $vollist) = @_;

    my $lvs = lvm_lvs ();

    foreach my $volid (@$vollist) {
	my ($storeid, $volname) = parse_volume_id ($volid);

	my $scfg = storage_config ($cfg, $storeid);

	if ($scfg->{type} eq 'lvm') {
	    my ($name) = parse_volname_lvm ($volname);

	    if ($lvs->{$scfg->{vgname}}->{$name}) {
		my $path = path ($cfg, $volid);
		my $cmd = ['/sbin/lvchange', '-aln', $path];
		eval { run_command ($cmd); };
		warn $@ if $@;
	    }
	}
    }
}

sub deactivate_storage {
    my ($cfg, $storeid) = @_;

    my $iscsi_sessions;

    my $scfg = storage_config ($cfg, $storeid);

    my $type = $scfg->{type};

    if ($type eq 'dir') {
	# nothing to do
    } elsif ($type eq 'nfs') {
	my $mountdata = read_proc_mounts();
	my $server = $scfg->{server};
	my $export = $scfg->{export};
	my $path = $scfg->{path};

	my $cmd = ['/bin/umount', $path];

	run_command ($cmd) if nfs_is_mounted ($server, $export, $path, $mountdata); 
    } elsif ($type eq 'lvm') {
	my $cmd = ['/sbin/vgchange', '-aln', $scfg->{vgname}];
	run_command ($cmd);
    } elsif ($type eq 'iscsi') {
	my $portal = $scfg->{portal};
	my $target = $scfg->{target};

	my $iscsi_sessions = iscsi_session_list();
	iscsi_logout ($target, $portal)
	    if defined ($iscsi_sessions->{$target});

    } else {
	die "implement me";
    }
}

sub storage_info { 
    my ($cfg) = @_;

    my $ids = $cfg->{ids};

    my $info = {};
    my $stypes = {};

    my $slist = [];
    foreach my $storeid (keys %$ids) {

	my $type = $ids->{$storeid}->{type};

	$info->{$storeid} = { 
	    type => $type,
	    total => 0, 
	    avail => 0, 
	    used => 0, 
	    free => 0, 
	    disable => defined ($ids->{$storeid}->{disable}),
	    active => 0,
	};

	next if $ids->{$storeid}->{disable};

	$stypes->{$type} = 1;

	push @$slist, $storeid;
    }

    eval { activate_storage_list ($cfg, $slist); };

    my $mountdata = $stypes->{nfs} ? read_proc_mounts() : '';
    my $iscsi_sessions = $stypes->{iscsi} ? iscsi_session_list() : {};

    foreach my $storeid (keys %$ids) {
	my $scfg = $ids->{$storeid};

	my $type = $scfg->{type};

	if ($type eq 'dir' || $type eq 'nfs') {

	    my $path = $scfg->{path};
	    
	    if ($type eq 'nfs') {
		my $server = $scfg->{server};
		my $export = $scfg->{export};

		next if !nfs_is_mounted ($server, $export, $path, $mountdata); 
	    }

	    my $res = df ($path);

	    next if !$res;

	    # available memory =  total memory - reserved memory
	    my $bavail = $res->{used}+$res->{bavail};

	    $info->{$storeid}->{total} = $res->{blocks}; 
	    $info->{$storeid}->{avail} = $bavail; 
	    $info->{$storeid}->{used} = $res->{used}; 
	    $info->{$storeid}->{free} = $bavail - $res->{used}; 
	    $info->{$storeid}->{active} = 1;

	} elsif ($type eq 'lvm') {

	    my $vgname = $scfg->{vgname};

	    my $total = 0;
	    my $free = 0;

	    my $res = lvm_vgs (); 

	    if (defined ($res->{$vgname})) {
		$total = $res->{$vgname}->{size};
		$free = $res->{$vgname}->{free};

		$info->{$storeid}->{total} = $total; 
		$info->{$storeid}->{avail} = $total; 
		$info->{$storeid}->{used} = $total - $free; 
		$info->{$storeid}->{free} = $free; 
		$info->{$storeid}->{active} = 1;
	    }

	} elsif ($type eq 'iscsi') {

	    $info->{$storeid}->{total} = 0; 
	    $info->{$storeid}->{avail} = 0; 
	    $info->{$storeid}->{used} = 0; 
	    $info->{$storeid}->{free} = 0; 
	    $info->{$storeid}->{active} = 
		defined ($iscsi_sessions->{$scfg->{target}});

	} else {
	    die "implement me";
	}
    }

    return $info;
}

sub resolv_server {
    my ($server) = @_;
    
    my $packed_ip = gethostbyname($server);
    if (defined $packed_ip) {
	return inet_ntoa($packed_ip);
    }
    return undef;
}

sub scan_nfs {
    my ($server_in) = @_;

    my $server;
    if (!($server = resolv_server ($server_in))) {
	die "unable to resolve address for server '${server_in}'\n";
    }

    my $cmd = ['/sbin/showmount',  '--no-headers', '--exports', $server];

    my $res = {};
    run_command ($cmd, outfunc => sub {
	my $line = shift;

	# note: howto handle white spaces in export path??
	if ($line =~ m!^(/\S+)\s+(.+)$!) {
	    $res->{$1} = $2;
	}
    });

    return $res;
}

sub resolv_portal {
    my ($portal) = @_;

    if ($portal =~ m/^([^:]+)(:(\d+))?$/) {
	my $server = $1;
	my $port = $3;

	my $packed_ip = gethostbyname($server);
	if (defined $packed_ip) {
	    $server = inet_ntoa($packed_ip);
	    $portal = $port ? "$server:$port" : $server;
	    return $portal;
	}
    }
    return undef;
}

sub scan_iscsi {
    my ($portal_in) = @_;

    my $portal;
    if (!($portal = resolv_portal ($portal_in))) {
	die "unable to parse/resolve portal address '${portal_in}'\n";
    }

    return iscsi_discovery ($portal);
}

sub storage_default_format {
    my ($cfg, $storeid) = @_;

    my $scfg = storage_config ($cfg, $storeid);

    my $def = $default_config->{$scfg->{type}};

    my $def_format = 'raw';
    my $valid_formats = [ $def_format ];

    if (defined ($def->{format})) {
	$def_format = $scfg->{format} || $def->{format}->[1];
	$valid_formats = [ sort keys %{$def->{format}->[0]} ];
    }
   
    return wantarray ? ($def_format, $valid_formats) : $def_format;
}

sub vgroup_is_used {
    my ($cfg, $vgname) = @_;

    foreach my $storeid (keys %{$cfg->{ids}}) {
	my $scfg = storage_config ($cfg, $storeid);
	if ($scfg->{type} eq 'lvm' && $scfg->{vgname} eq $vgname) {
	    return 1;
	}
    }

    return undef;
}

sub target_is_used {
    my ($cfg, $target) = @_;

    foreach my $storeid (keys %{$cfg->{ids}}) {
	my $scfg = storage_config ($cfg, $storeid);
	if ($scfg->{type} eq 'iscsi' && $scfg->{target} eq $target) {
	    return 1;
	}
    }

    return undef;
}

sub volume_is_used {
    my ($cfg, $volid) = @_;

    foreach my $storeid (keys %{$cfg->{ids}}) {
	my $scfg = storage_config ($cfg, $storeid);
	if ($scfg->{base} && $scfg->{base} eq $volid) {
	    return 1;
	}
    }

    return undef;
}

sub storage_is_used {
    my ($cfg, $storeid) = @_;

    foreach my $sid (keys %{$cfg->{ids}}) {
	my $scfg = storage_config ($cfg, $sid);
	next if !$scfg->{base};
	my ($st) = parse_volume_id ($scfg->{base});
	return 1 if $st && $st eq $storeid;
    }

    return undef;
}


sub storage_add {
    my ($storeid, $type, $param) = @_;

    my $opts = parse_options ($storeid, $type, $param, 1);

    lock_config (sub {

	my $cfg = load_config();

	if (my $scfg = storage_config ($cfg, $storeid, 1)) {
	    die "storage ID '$storeid' already defined\n";
	}

	$cfg->{ids}->{$storeid} = $opts;

	if ($type eq 'lvm' && $opts->{base}) {

	    my ($baseid, $volname) = parse_volume_id ($opts->{base});

	    my $basecfg = storage_config ($cfg, $baseid, 1);
	    die "base storage ID '$baseid' does not exist\n" if !$basecfg;
       
	    # we only support iscsi for now
	    if (!($basecfg->{type} eq 'iscsi')) {
		die "unsupported base type '$basecfg->{type}'";
	    }

	    my $path = path ($cfg, $opts->{base});

	    lvm_create_volume_group ($path, $opts->{vgname}, $opts->{shared});
	}
	
	activate_storage ($cfg, $storeid) if !$opts->{disable};

	write_config($cfg);
    });

    die $@ if $@;
}

sub storage_set {
    my ($storeid, $param, $digest) = @_;

    lock_config (sub {

	my $cfg = load_config();

	__assert_if_modified ($cfg, $digest);

	my $scfg = storage_config ($cfg, $storeid);

	my $opts = parse_options ($storeid, $scfg->{type}, $param);

	foreach my $k (%$opts) {
	    $scfg->{$k} = $opts->{$k};
	}
	
	write_config ($cfg);
    });

    die $@ if $@;
}

sub storage_remove {
    my ($storeid, $digest) = @_;

    lock_config (sub {

	my $cfg = load_config();

	__assert_if_modified ($cfg, $digest);

	die "can't remove storage - storage is used as base of another storage\n"
	    if storage_is_used ($cfg, $storeid);

	deactivate_storage ($cfg, $storeid);

	delete ($cfg->{ids}->{$storeid});

	write_config ($cfg);
    });
 
    die $@ if $@;
}

sub storage_enable {
    my ($storeid, $digest) = @_;

    lock_config (sub {

	my $cfg = load_config();

	__assert_if_modified ($cfg, $digest);

	my $scfg = storage_config ($cfg, $storeid);

	delete $scfg->{disable};

	activate_storage ($cfg, $storeid);	

	write_config ($cfg);
    });

    die $@ if $@;
}

sub storage_disable {
    my ($storeid, $digest) = @_;

    lock_config (sub {

	my $cfg = load_config();

	__assert_if_modified ($cfg, $digest);

	my $scfg = storage_config ($cfg, $storeid);

	$scfg->{disable} = 1;

	deactivate_storage ($cfg, $storeid);	

	write_config ($cfg);
    });

    die $@ if $@;
}

sub foreach_volid {
    my ($list, $func) = @_;

    return if !$list;

    foreach my $sid (keys %$list) {
	foreach my $info (@{$list->{$sid}}) {
	    my $volname = $info->{volname};
	    my $volid = "$sid:$volname";
	    &$func ($volid, $sid, $volname, $info);
	}
    }
}

1;
